from __future__ import annotations

from sqlalchemy import ForeignKey
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

from xladmin import AdminConfig, AdminFieldConfig, AdminModelConfig, AdminRegistry, ModelConfig
from xladmin.introspection import get_model_meta, get_visible_list_fields


class Base(DeclarativeBase):
    pass


class DemoModel(Base):
    __tablename__ = "demo_model"

    id: Mapped[int] = mapped_column(primary_key=True)


class OrderedParentModel(Base):
    __tablename__ = "ordered_parent_model"

    id: Mapped[int] = mapped_column(primary_key=True)
    title: Mapped[str] = mapped_column()
    children: Mapped[list[OrderedChildModel]] = relationship(back_populates="parent")


class OrderedChildModel(Base):
    __tablename__ = "ordered_child_model"

    id: Mapped[int] = mapped_column(primary_key=True)
    parent_id: Mapped[int] = mapped_column(ForeignKey("ordered_parent_model.id"))
    parent: Mapped[OrderedParentModel] = relationship(back_populates="children")


def test_registry_stores_model_configs() -> None:
    registry = AdminRegistry(AdminModelConfig(model=DemoModel, slug="demo", title="Demo"))

    config = registry.get("demo")

    assert config.model is DemoModel
    assert config.title == "Demo"


def test_registry_builds_defaults_from_admin_config() -> None:
    registry = AdminRegistry(config=AdminConfig(models=(ModelConfig(model=DemoModel),)))

    config = registry.list()[0]

    assert config.model is DemoModel
    assert config.slug == "demo-model"
    assert config.title == "Demo Model"
    assert config.ordering == ("-id",)


def test_visible_list_fields_follow_model_declaration_order() -> None:
    config = ModelConfig(model=OrderedParentModel)

    assert get_visible_list_fields(config) == ["id", "title", "children"]


def test_model_meta_includes_display_kind() -> None:
    config = ModelConfig(
        model=DemoModel,
        fields={"id": AdminFieldConfig(display_kind="image", image_url_prefix="/media")},
    )

    meta = get_model_meta(config)
    field_meta = next(field for field in meta["fields"] if field["name"] == "id")

    assert field_meta["display_kind"] == "image"
    assert field_meta["image_url_prefix"] == "/media"



from __future__ import annotations

import builtins
import re
from typing import Any

from xladmin.config import AdminConfig, ModelConfig, ModelsBlockConfig
from xladmin.i18n import normalize_locale


class Registry:
    """Runtime registry for configured admin models."""

    def __init__(
            self,
            *configs: ModelConfig,
            config: AdminConfig | None = None,
            model_blocks: tuple[ModelsBlockConfig, ...] = (),
    ) -> None:
        self._configs_by_slug: dict[str, ModelConfig] = {}
        self._model_blocks = config.models_blocks if config is not None else model_blocks
        self.locale = normalize_locale(config.locale if config is not None else None)

        combined_configs = list(config.models) if config is not None else []
        combined_configs.extend(configs)
        for model_config in combined_configs:
            self.register(model_config)

    def register(self, config: ModelConfig) -> None:
        normalized_config = normalize_config(config)
        slug = normalized_config.slug
        if slug is None:
            raise ValueError("Model slug must be resolved before registration.")
        if slug in self._configs_by_slug:
            raise ValueError(f"Model slug '{slug}' is already registered.")
        self._configs_by_slug[slug] = normalized_config

    def get(self, slug: str) -> ModelConfig:
        try:
            return self._configs_by_slug[slug]
        except KeyError as exc:
            raise KeyError(f"Admin model '{slug}' is not registered.") from exc

    def find_by_model(self, model: type[Any]) -> ModelConfig | None:
        for config in self._configs_by_slug.values():
            if config.model is model:
                return config
        return None

    def list(self) -> builtins.list[ModelConfig]:
        return list(self._configs_by_slug.values())

    def list_model_blocks(self) -> builtins.list[ModelsBlockConfig]:
        return list(self._model_blocks)

    def resolve_model_blocks(self) -> builtins.list[tuple[ModelsBlockConfig, builtins.list[ModelConfig]]]:
        resolved_blocks: builtins.list[tuple[ModelsBlockConfig, builtins.list[ModelConfig]]] = []
        for block in self._model_blocks:
            block_models: builtins.list[ModelConfig] = []
            references = block.models or block.model_slugs
            for item in references:
                if isinstance(item, str):
                    block_models.append(self.get(item))
                    continue

                model_config = self.find_by_model(item)
                if model_config is None:
                    raise KeyError(f"Admin model for '{item.__name__}' is not registered.")
                block_models.append(model_config)

            resolved_blocks.append((block, block_models))
        return resolved_blocks


def build_registry(registry_or_config: Registry | AdminConfig) -> Registry:
    """Normalize config or registry into a runtime registry."""

    if isinstance(registry_or_config, Registry):
        return registry_or_config
    if isinstance(registry_or_config, AdminConfig):
        return Registry(config=registry_or_config)
    raise TypeError("Admin registry must be Registry or AdminConfig.")


def normalize_config(config: ModelConfig) -> ModelConfig:
    """Fill base defaults for minimal `ModelConfig(model=SomeORM)`."""

    if config.slug is None:
        config.slug = get_default_model_slug(config.model)
    if config.title is None:
        config.title = get_default_model_title(config.model)
    if not config.search_fields:
        config.search_fields = get_default_search_fields(config.model)
    if not config.ordering:
        config.ordering = get_default_ordering(config.model)
    return config


def get_default_model_slug(model: type[Any]) -> str:
    module_parts = model.__module__.split(".")
    try:
        modules_index = module_parts.index("modules")
    except ValueError:
        prefix = "model"
    else:
        try:
            models_index = module_parts.index("models", modules_index + 1)
        except ValueError:
            relevant_parts = module_parts[modules_index + 1:]
        else:
            relevant_parts = module_parts[modules_index + 1:models_index]
        prefix = "-".join(relevant_parts) or "model"

    base_name = to_kebab_case(get_base_model_name(model))
    return f"{prefix}-{base_name}" if prefix != "model" else base_name


def get_default_model_title(model: type[Any]) -> str:
    return to_title_case(to_snake_case(get_base_model_name(model)))


def get_default_search_fields(model: type[Any]) -> tuple[str, ...]:
    table = getattr(model, "__table__", None)
    if table is None:
        return ()

    search_fields: list[str] = []
    for field_name in ("name", "title", "slug", "email", "username", "code"):
        if field_name in table.columns:
            search_fields.append(field_name)
    return tuple(search_fields)


def get_default_ordering(model: type[Any]) -> tuple[str, ...]:
    table = getattr(model, "__table__", None)
    if table is None:
        return ()

    for field_name in ("created_at", "updated_at", "date_joined", "id"):
        if field_name in table.columns:
            return (f"-{field_name}",)
    return ()


def get_base_model_name(model: type[Any]) -> str:
    if model.__name__.endswith("ORM"):
        return model.__name__[:-3]
    return model.__name__


def to_snake_case(value: str) -> str:
    normalized_value = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", value)
    normalized_value = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", normalized_value)
    return normalized_value.lower()


def to_kebab_case(value: str) -> str:
    return to_snake_case(value).replace("_", "-")


def to_title_case(value: str) -> str:
    return value.replace("_", " ").strip().title()


AdminRegistry = Registry
build_admin_registry = build_registry
normalize_model_config = normalize_config



from xladmin.config import (
    AdminBulkActionConfig,
    AdminConfig,
    AdminFieldConfig,
    AdminHTTPConfig,
    AdminListFilterConfig,
    AdminListFilterOptionConfig,
    AdminLocale,
    AdminModelConfig,
    AdminModelsBlockConfig,
    AdminObjectActionConfig,
    BulkActionConfig,
    FieldConfig,
    FieldDisplayKind,
    FieldInputKind,
    HttpConfig,
    ListFilterConfig,
    ListFilterInputKind,
    ListFilterOptionConfig,
    ModelConfig,
    ModelsBlock,
    ModelsBlockConfig,
    ObjectActionConfig,
)
from xladmin.registry import (
    AdminRegistry,
    Registry,
    build_admin_registry,
    build_registry,
    normalize_config,
    normalize_model_config,
)
from xladmin.router import create_admin_router, create_router

__all__ = (
    "AdminBulkActionConfig",
    "AdminConfig",
    "AdminFieldConfig",
    "AdminHTTPConfig",
    "AdminListFilterConfig",
    "AdminListFilterOptionConfig",
    "AdminModelConfig",
    "AdminModelsBlockConfig",
    "AdminObjectActionConfig",
    "AdminRegistry",
    "BulkActionConfig",
    "FieldConfig",
    "FieldDisplayKind",
    "FieldInputKind",
    "HttpConfig",
    "AdminLocale",
    "ListFilterConfig",
    "ListFilterInputKind",
    "ListFilterOptionConfig",
    "ModelConfig",
    "ModelsBlock",
    "ModelsBlockConfig",
    "ObjectActionConfig",
    "Registry",
    "build_admin_registry",
    "build_registry",
    "create_admin_router",
    "create_router",
    "normalize_config",
    "normalize_model_config",
)



from __future__ import annotations

from datetime import date, datetime
from decimal import Decimal
from typing import Any
from uuid import UUID

from sqlalchemy import inspect as sa_inspect
from sqlalchemy.orm.attributes import NO_VALUE

from xladmin.config import AdminModelConfig
from xladmin.introspection import (
    get_display_value,
    get_field_value,
    get_pk_field_name,
    get_pk_value,
    get_relationship_names,
    get_visible_detail_fields,
    get_visible_list_fields,
)


def serialize_model_instance(
        config: AdminModelConfig,
        instance: Any,
        *,
        mode: str,
) -> dict[str, Any]:
    fields = get_visible_list_fields(config) if mode == "list" else get_visible_detail_fields(config)
    payload = {field_name: serialize_scalar(get_field_value(config, instance, field_name)) for field_name in fields}
    payload[get_pk_field_name(config)] = serialize_scalar(get_pk_value(config, instance))
    payload["_display"] = get_display_value(config, instance)
    if mode == "detail":
        payload["_relations"] = serialize_relations(config, instance)
    return payload


def serialize_relations(config: AdminModelConfig, instance: Any) -> dict[str, Any]:
    mapper = sa_inspect(config.model)
    instance_state = sa_inspect(instance)
    payload: dict[str, Any] = {}
    for relation_name in get_relationship_names(config):
        relationship = mapper.relationships[relation_name]
        if instance_state.attrs[relation_name].loaded_value is NO_VALUE:
            payload[relation_name] = [] if relationship.uselist else None
            continue
        value = getattr(instance, relation_name)
        if value is None:
            payload[relation_name] = None
            continue
        if relationship.uselist:
            items = list(value)
            payload[relation_name] = {
                "count": len(items),
                "items": [_serialize_related_item(item) for item in items[:5]],
            }
            continue
        payload[relation_name] = _serialize_related_item(value)
    return payload


def serialize_scalar(value: Any) -> Any:
    if isinstance(value, Decimal):
        return str(value)
    if isinstance(value, (datetime, date)):
        return value.isoformat()
    if isinstance(value, UUID):
        return str(value)
    return value


def _serialize_related_item(instance: Any) -> dict[str, Any]:
    mapper = sa_inspect(type(instance))
    pk_column = mapper.primary_key[0]
    pk_name = pk_column.key
    label = None
    for candidate in ("name", "title", "username", "slug", "email", "code_name"):
        if hasattr(instance, candidate):
            raw = getattr(instance, candidate)
            if raw is not None:
                label = str(raw)
                break
    return {
        "id": serialize_scalar(getattr(instance, pk_name)),
        "label": label or f"{type(instance).__name__} #{getattr(instance, pk_name)}",
    }



from __future__ import annotations

from typing import Any, Literal

from sqlalchemy import inspect as sa_inspect
from sqlalchemy import or_, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import RelationshipDirection

from xladmin.config import ModelConfig
from xladmin.introspection import get_display_value, get_pk_value
from xladmin.registry import Registry
from xladmin.serializer import serialize_scalar

DeleteEffect = Literal["delete", "protect", "set-null"]


async def build_delete_plan(
        session: AsyncSession,
        registry: Registry,
        model_config: ModelConfig,
        items: list[Any],
) -> tuple[dict[str, Any], list[Any], list[tuple[Any, list[str]]]]:
    ordered_items: list[Any] = []
    ordered_keys: set[tuple[type[Any], Any]] = set()
    set_null_items: list[tuple[Any, list[str]]] = []
    set_null_keys: set[tuple[type[Any], Any]] = set()
    roots = [
        await _build_preview_node(
            session=session,
            registry=registry,
            model_config=model_config,
            item=item,
            relation_name=None,
            path_seen=set(),
            ordered_items=ordered_items,
            ordered_keys=ordered_keys,
            set_null_items=set_null_items,
            set_null_keys=set_null_keys,
            effect="delete",
        )
        for item in items
    ]
    counts = _count_effects(roots)

    return {
        "can_delete": counts["protect"] == 0,
        "summary": {
            "roots": len(roots),
            "delete": counts["delete"],
            "protect": counts["protect"],
            "set_null": counts["set-null"],
            "total": len(roots) + counts["delete"] + counts["protect"] + counts["set-null"],
        },
        "roots": roots,
    }, ordered_items, set_null_items


async def build_delete_preview(
        session: AsyncSession,
        registry: Registry,
        model_config: ModelConfig,
        items: list[Any],
) -> dict[str, Any]:
    preview, _ordered_items, _set_null_items = await build_delete_plan(session, registry, model_config, items)
    return preview


async def _build_preview_node(
        session: AsyncSession,
        registry: Registry,
        model_config: ModelConfig,
        item: Any,
        relation_name: str | None,
        path_seen: set[tuple[type[Any], Any]],
        ordered_items: list[Any],
        ordered_keys: set[tuple[type[Any], Any]],
        set_null_items: list[tuple[Any, list[str]]],
        set_null_keys: set[tuple[type[Any], Any]],
        effect: DeleteEffect,
) -> dict[str, Any]:
    item_id = get_pk_value(model_config, item)
    node_key = (type(item), item_id)
    if node_key in path_seen:
        return _build_node_payload(
            model_slug=model_config.slug,
            model_title=_get_model_title(model_config),
            relation_name=relation_name,
            item_id=item_id,
            label=get_display_value(model_config, item),
            children=[],
            effect=effect,
        )

    next_path_seen = set(path_seen)
    next_path_seen.add(node_key)

    children: list[dict[str, Any]] = []

    for dependent_relation in _iter_dependent_relations(model_config.model):
        related_config = registry.find_by_model(dependent_relation.model)
        related_items = list((await session.execute(_build_dependent_query(item_id, dependent_relation))).scalars())
        for related_item in related_items:
            if related_config is None:
                children.append(
                    _build_node_payload(
                        model_slug=None,
                        model_title=dependent_relation.model.__name__,
                        relation_name=dependent_relation.relation_name,
                        item_id=_get_raw_pk_value(related_item),
                        label=_get_fallback_label(related_item),
                        children=[],
                        effect=dependent_relation.effect,
                    ),
                )
                continue

            if dependent_relation.effect == "set-null":
                related_key = (type(related_item), _get_raw_pk_value(related_item))
                if related_key not in set_null_keys:
                    set_null_items.append((related_item, dependent_relation.attribute_names))
                    set_null_keys.add(related_key)
                children.append(
                    _build_node_payload(
                        model_slug=related_config.slug,
                        model_title=_get_model_title(related_config),
                        relation_name=dependent_relation.relation_name,
                        item_id=get_pk_value(related_config, related_item),
                        label=get_display_value(related_config, related_item),
                        children=[],
                        effect="set-null",
                    ),
                )
                continue

            children.append(
                await _build_preview_node(
                    session=session,
                    registry=registry,
                    model_config=related_config,
                    item=related_item,
                    relation_name=dependent_relation.relation_name,
                    path_seen=next_path_seen,
                    ordered_items=ordered_items,
                    ordered_keys=ordered_keys,
                    set_null_items=set_null_items,
                    set_null_keys=set_null_keys,
                    effect=dependent_relation.effect,
                ),
            )

    if effect == "delete" and node_key not in ordered_keys:
        ordered_items.append(item)
        ordered_keys.add(node_key)

    return _build_node_payload(
        model_slug=model_config.slug,
        model_title=_get_model_title(model_config),
        relation_name=relation_name,
        item_id=item_id,
        label=get_display_value(model_config, item),
        children=children,
        effect=effect,
    )


def _build_dependent_query(item_id: Any, dependent_relation: _DependentRelation):
    return select(dependent_relation.model).where(
        or_(*[
            getattr(dependent_relation.model, attribute_name) == item_id
            for attribute_name in dependent_relation.attribute_names
        ]),
    )


def _iter_dependent_relations(model: type[Any]) -> list[_DependentRelation]:
    mapper = sa_inspect(model)
    pk_column = mapper.primary_key[0]
    relations: list[_DependentRelation] = []

    for candidate_mapper in mapper.registry.mappers:
        candidate_model = candidate_mapper.class_
        if candidate_model is model:
            continue

        columns = [
            column
            for column in candidate_mapper.columns
            if any(
                foreign_key.column.table is pk_column.table and foreign_key.column.key == pk_column.key
                for foreign_key in column.foreign_keys
            )
        ]
        if not columns:
            continue
        attribute_names = [_get_attribute_name_for_column(candidate_mapper, column) for column in columns]

        relations.append(
            _DependentRelation(
                model=candidate_model,
                relation_name=", ".join(column.key for column in columns),
                columns=columns,
                attribute_names=attribute_names,
                effect=_resolve_dependent_effect(model, candidate_model, columns),
            ),
        )

    return relations


def _resolve_dependent_effect(
        parent_model: type[Any],
        candidate_model: type[Any],
        columns: list[Any],
) -> DeleteEffect:
    relationship_effect = _get_relationship_effect(parent_model, candidate_model, columns)
    if relationship_effect is not None:
        return relationship_effect
    if any(_column_has_ondelete(column, "CASCADE") for column in columns):
        return "delete"
    if any(_column_has_ondelete(column, "SET NULL") for column in columns):
        return "set-null"
    return "protect"


def _get_relationship_effect(
        parent_model: type[Any],
        candidate_model: type[Any],
        columns: list[Any],
) -> DeleteEffect | None:
    parent_mapper = sa_inspect(parent_model)
    column_keys = {column.key for column in columns}

    for relationship in parent_mapper.relationships:
        if relationship.mapper.class_ is not candidate_model:
            continue
        if relationship.secondary is not None or relationship.viewonly:
            continue
        if relationship.direction is not RelationshipDirection.ONETOMANY:
            continue

        relationship_column_keys = {remote.key for _local, remote in relationship.local_remote_pairs}
        if relationship_column_keys and not relationship_column_keys.issubset(column_keys):
            continue
        if _relationship_cascades_delete(relationship.cascade):
            return "delete"

    return None


def _relationship_cascades_delete(cascade: Any) -> bool:
    return bool(getattr(cascade, "delete", False) or getattr(cascade, "delete_orphan", False))


def _column_has_ondelete(column: Any, value: str) -> bool:
    return any(
        isinstance(getattr(foreign_key, "ondelete", None), str) and foreign_key.ondelete.upper() == value
        for foreign_key in column.foreign_keys
    )


def _get_attribute_name_for_column(candidate_mapper: Any, column: Any) -> str:
    return candidate_mapper.get_property_by_column(column).key


def _count_effects(roots: list[dict[str, Any]]) -> dict[DeleteEffect, int]:
    counts: dict[DeleteEffect, int] = {"delete": 0, "protect": 0, "set-null": 0}
    for root in roots:
        for child in root.get("children", []):
            _accumulate_effects(child, counts)
    return counts


def _accumulate_effects(node: dict[str, Any], counts: dict[DeleteEffect, int]) -> None:
    effect = node.get("effect", "delete")
    counts[effect] += 1
    for child in node.get("children", []):
        _accumulate_effects(child, counts)


def _build_node_payload(
        *,
        model_slug: str | None,
        model_title: str,
        relation_name: str | None,
        item_id: Any,
        label: str,
        children: list[dict[str, Any]],
        effect: DeleteEffect,
) -> dict[str, Any]:
    return {
        "model_slug": model_slug,
        "model_title": model_title,
        "relation_name": relation_name,
        "id": serialize_scalar(item_id),
        "label": label,
        "effect": effect,
        "children": children,
    }


def _get_raw_pk_value(item: Any) -> Any:
    mapper = sa_inspect(type(item))
    pk_column = mapper.primary_key[0]
    return getattr(item, pk_column.key)


def _get_fallback_label(item: Any) -> str:
    item_id = serialize_scalar(_get_raw_pk_value(item))
    for candidate in ("name", "title", "username", "slug", "email", "code_name"):
        if hasattr(item, candidate):
            raw_value = getattr(item, candidate)
            if raw_value is not None:
                return str(raw_value)
    return f"{type(item).__name__} #{item_id}"


def _get_model_title(model_config: ModelConfig) -> str:
    return model_config.title or model_config.slug or model_config.model.__name__


class _DependentRelation:
    def __init__(
            self,
            *,
            model: type[Any],
            relation_name: str,
            columns: list[Any],
            attribute_names: list[str],
            effect: DeleteEffect,
    ) -> None:
        self.model = model
        self.relation_name = relation_name
        self.columns = columns
        self.attribute_names = attribute_names
        self.effect = effect



from __future__ import annotations

from typing import Any, cast

from xladmin.config import AdminLocale

SUPPORTED_LOCALES: set[AdminLocale] = {"ru", "en"}
DEFAULT_LOCALE: AdminLocale = "ru"

_MESSAGES: dict[str, dict[str, str]] = {
    "ru": {
        "delete": "Удалить",
        "yes": "Да",
        "no": "Нет",
        "forbidden": "Доступ запрещён.",
        "admin_model_not_found": "Модель админки не найдена.",
        "object_not_found": "Объект не найден.",
        "object_action_not_found": "Действие объекта не найдено.",
        "bulk_action_not_found": "Массовое действие не найдено.",
        "list_filter_not_found": "Фильтр списка не найден.",
        "field_has_no_relation_choices": "У поля нет связанных вариантов выбора.",
        "related_model_not_found": "Связанная модель не найдена.",
        "delete_blocked": "Удаление заблокировано связанными объектами.",
    },
    "en": {
        "delete": "Delete",
        "yes": "Yes",
        "no": "No",
        "forbidden": "Access denied.",
        "admin_model_not_found": "Admin model was not found.",
        "object_not_found": "Object was not found.",
        "object_action_not_found": "Object action was not found.",
        "bulk_action_not_found": "Bulk action was not found.",
        "list_filter_not_found": "List filter was not found.",
        "field_has_no_relation_choices": "The field has no related choices.",
        "related_model_not_found": "Related model was not found.",
        "delete_blocked": "Deletion is blocked by related objects.",
    },
}


def normalize_locale(locale: str | None) -> AdminLocale:
    if locale == "ru" or locale == "en":
        return cast(AdminLocale, locale)
    return DEFAULT_LOCALE


def translate(locale: str | None, key: str, **kwargs: Any) -> str:
    normalized_locale = normalize_locale(locale)
    template = _MESSAGES.get(normalized_locale, _MESSAGES[DEFAULT_LOCALE]).get(key, key)
    if kwargs:
        return template.format(**kwargs)
    return template



from __future__ import annotations

from datetime import date, datetime
from decimal import Decimal
from typing import Any
from uuid import UUID

from sqlalchemy import inspect as sa_inspect
from sqlalchemy.orm import ColumnProperty, RelationshipProperty

from xladmin.config import AdminModelConfig
from xladmin.i18n import normalize_locale, translate
from xladmin.registry import AdminRegistry


def get_mapper_field_names(config: AdminModelConfig) -> list[str]:
    """
    Возвращает имена ORM-атрибутов в их естественном порядке объявления в модели.

    Это основной дефолтный порядок для detail/list/create/update представлений,
    если проект не переопределил его явно через конфиг.
    """

    mapper = sa_inspect(config.model)
    ordered_names: list[str] = []
    declared_names = getattr(config.model, "__annotations__", {})

    for field_name in declared_names:
        if field_name not in mapper.attrs:
            continue
        prop = mapper.attrs[field_name]
        if isinstance(prop, ColumnProperty | RelationshipProperty):
            ordered_names.append(prop.key)

    for prop in mapper.attrs:
        if not isinstance(prop, ColumnProperty | RelationshipProperty):
            continue
        if prop.key not in ordered_names:
            ordered_names.append(prop.key)

    return ordered_names


def get_column_names(config: AdminModelConfig) -> list[str]:
    mapper = sa_inspect(config.model)
    return [prop.key for prop in mapper.attrs if isinstance(prop, ColumnProperty)]


def get_relationship_names(config: AdminModelConfig) -> list[str]:
    mapper = sa_inspect(config.model)
    return [prop.key for prop in mapper.attrs if isinstance(prop, RelationshipProperty)]


def get_all_field_names(config: AdminModelConfig) -> list[str]:
    """
    Возвращает полный список полей модели для админки.

    Помимо ORM-колонок сюда входят и виртуальные поля, которые клиент библиотеки
    описал в `config.fields`, например `new_password`.
    """

    ordered_names = list(get_mapper_field_names(config))
    configured_names = _get_configured_field_names(config)
    for field_name in config.fields:
        if field_name not in ordered_names:
            ordered_names.append(field_name)
    for field_name in configured_names:
        if field_name not in ordered_names:
            ordered_names.append(field_name)
    return ordered_names


def get_visible_list_fields(config: AdminModelConfig) -> list[str]:
    if config.list_display is not None:
        return list(config.list_display)
    if config.list_fields is not None:
        return list(config.list_fields)
    return [name for name in get_all_field_names(config) if not config.get_field_config(name).hidden_in_list]


def get_visible_detail_fields(config: AdminModelConfig) -> list[str]:
    if config.detail_fields is not None:
        return list(config.detail_fields)
    return [name for name in get_all_field_names(config) if not config.get_field_config(name).hidden_in_detail]


def get_create_fields(config: AdminModelConfig) -> list[str]:
    if config.create_fields is not None:
        return list(config.create_fields)
    mapper = sa_inspect(config.model)
    fields: list[str] = []
    column_names = set(get_column_names(config))
    relationship_names = set(get_relationship_names(config))
    for name in get_all_field_names(config):
        field_config = config.get_field_config(name)
        if _is_read_only(config, name) or field_config.hidden_in_form:
            continue
        if name in column_names:
            column = mapper.columns[name]
            if name == get_pk_field_name(config) and _pk_is_generated(column):
                continue
        elif name in relationship_names:
            pass
        elif field_config.value_setter is None:
            continue
        fields.append(name)
    return fields


def get_update_fields(config: AdminModelConfig) -> list[str]:
    if config.update_fields is not None:
        return list(config.update_fields)
    relationship_names = set(get_relationship_names(config))
    return [
        name for name in get_all_field_names(config)
        if (
                name != get_pk_field_name(config)
                and not _is_read_only(config, name)
                and not config.get_field_config(name).hidden_in_form
                and (
                        name in get_column_names(config)
                        or name in relationship_names
                        or config.get_field_config(name).value_setter is not None
                )
        )
    ]


def get_pk_field_name(config: AdminModelConfig) -> str:
    mapper = sa_inspect(config.model)
    if hasattr(config.model, config.pk_field):
        return config.pk_field
    return mapper.primary_key[0].key


def get_pk_value(config: AdminModelConfig, instance: Any) -> Any:
    return getattr(instance, get_pk_field_name(config))


def get_display_value(config: AdminModelConfig, instance: Any) -> str:
    if config.display_field:
        value = getattr(instance, config.display_field, None)
        if value is not None:
            return str(value)
    pk_value = get_pk_value(config, instance)
    return f"{config.title} #{pk_value}"


def get_model_meta(config: AdminModelConfig, *, locale: str = "ru") -> dict[str, Any]:
    normalized_locale = normalize_locale(locale)
    mapper = sa_inspect(config.model)
    relation_names = set(get_relationship_names(config))
    column_names = set(get_column_names(config))
    fields: list[dict[str, Any]] = []

    for name in get_all_field_names(config):
        field_config = config.get_field_config(name)
        column = mapper.columns[name] if name in column_names else None
        fields.append(
            {
                "name": name,
                "label": field_config.label or name,
                "help_text": field_config.help_text,
                "width_px": field_config.width_px,
                "display_kind": field_config.display_kind or "text",
                "image_url_prefix": field_config.image_url_prefix,
                "nullable": bool(column.nullable) if column is not None else True,
                "read_only": _is_read_only(config, name),
                "hidden_in_list": field_config.hidden_in_list,
                "hidden_in_detail": field_config.hidden_in_detail,
                "hidden_in_form": field_config.hidden_in_form,
                "type": _get_field_type(config, name, column),
                "input_kind": _get_input_kind(config, name, column),
                "is_primary_key": bool(column.primary_key) if column is not None else False,
                "is_virtual": column is None,
                "has_choices": _has_relation_choice_endpoint(config, name, relation_names),
                "is_relation": name in relation_names or (column is not None and bool(column.foreign_keys)),
                "is_relation_many": name in relation_names and bool(mapper.relationships[name].uselist),
                "is_sortable": get_sort_column_name(config, name) is not None,
            },
        )

    return {
        "locale": normalized_locale,
        "slug": config.slug,
        "title": config.title,
        "description": config.description,
        "pk_field": get_pk_field_name(config),
        "display_field": config.display_field,
        "page_size": config.page_size,
        "list_filters": [
            {
                "slug": list_filter.slug,
                "label": list_filter.label,
                "group": list_filter.group,
                "field_name": list_filter.field_name,
                "input_kind": _get_list_filter_input_kind(config, list_filter),
                "placeholder": list_filter.placeholder,
                "has_choices": _list_filter_has_choices(config, list_filter),
                "options": _get_list_filter_options(normalized_locale, config, list_filter),
            }
            for list_filter in config.list_filters
        ],
        "list_fields": get_visible_list_fields(config),
        "detail_fields": get_visible_detail_fields(config),
        "create_fields": get_create_fields(config),
        "update_fields": get_update_fields(config),
        "bulk_actions": [
            {"slug": "delete", "label": translate(normalized_locale, "delete")}, *[
                {"slug": action.slug, "label": action.label}
                for action in config.bulk_actions
            ]
        ],
        "object_actions": [
            {"slug": action.slug, "label": action.label}
            for action in config.object_actions
        ],
        "fields": fields,
    }


def get_model_blocks_meta(registry: AdminRegistry) -> list[dict[str, Any]]:
    return [
        {
            "slug": block.slug,
            "title": block.title,
            "description": block.description,
            "color": block.color,
            "collapsible": block.collapsible,
            "default_expanded": block.default_expanded,
            "models": [get_model_meta(model_config, locale=registry.locale) for model_config in model_configs],
        }
        for block, model_configs in registry.resolve_model_blocks()
    ]


def convert_value_for_column(column, value: Any) -> Any:
    if value is None:
        return None

    try:
        python_type = column.type.python_type
    except NotImplementedError:
        python_type = None

    if python_type is None:
        return value
    if python_type is bool:
        return bool(value)
    if python_type is int:
        return int(value)
    if python_type is float:
        return float(value)
    if python_type is Decimal:
        return Decimal(str(value))
    if python_type is date:
        return date.fromisoformat(str(value))
    if python_type is datetime:
        return datetime.fromisoformat(str(value))
    if python_type is UUID:
        return UUID(str(value))
    if python_type is str:
        return str(value)
    return value


def get_field_value(config: AdminModelConfig, instance: Any, field_name: str) -> Any:
    field_config = config.get_field_config(field_name)
    if field_config.value_getter is not None:
        return field_config.value_getter(instance)

    mapper = sa_inspect(config.model)
    if field_name in mapper.relationships:
        relationship = mapper.relationships[field_name]
        related_value = getattr(instance, field_name, None)
        if relationship.uselist:
            return [_serialize_related_pk(item) for item in list(related_value or [])]
        if related_value is None:
            return None
        return _serialize_related_pk(related_value)

    return getattr(instance, field_name, None)


def get_sort_column_name(config: AdminModelConfig, field_name: str) -> str | None:
    field_config = config.get_field_config(field_name)
    if field_config.ordering_field is not None:
        return field_config.ordering_field

    mapper = sa_inspect(config.model)
    if field_name in mapper.columns:
        return field_name

    if field_name in mapper.relationships and not mapper.relationships[field_name].uselist:
        relationship = mapper.relationships[field_name]
        local_columns = list(relationship.local_columns)
        if local_columns:
            return local_columns[0].key

    return None


def _is_read_only(config: AdminModelConfig, field_name: str) -> bool:
    return config.get_field_config(field_name).read_only


def _pk_is_generated(column) -> bool:
    if not bool(column.primary_key):
        return False
    if column.autoincrement is True:
        return True
    if column.default is not None or column.server_default is not None:
        return True
    return False


def _has_relation_choice_endpoint(
        config: AdminModelConfig,
        field_name: str,
        relation_names: set[str],
) -> bool:
    field_config = config.get_field_config(field_name)
    if field_config.relation_model is not None:
        return True
    if field_name in relation_names:
        return True
    mapper = sa_inspect(config.model)
    if field_name not in mapper.columns:
        return False
    column = mapper.columns[field_name]
    return bool(column.foreign_keys)


def _get_field_type(config: AdminModelConfig, field_name: str, column: Any | None) -> str:
    field_config = config.get_field_config(field_name)
    if field_config.input_kind is not None:
        return field_config.input_kind
    mapper = sa_inspect(config.model)
    if field_name in mapper.relationships:
        return "relationship"
    if column is None:
        return "virtual"
    return str(column.type)


def _get_input_kind(config: AdminModelConfig, field_name: str, column: Any | None) -> str:
    field_config = config.get_field_config(field_name)
    if field_config.input_kind is not None:
        return field_config.input_kind
    mapper = sa_inspect(config.model)
    if field_name in mapper.relationships:
        return "relation-multiple" if mapper.relationships[field_name].uselist else "relation"
    if column is None:
        return "text"

    if "json" in str(column.type).lower():
        return "json"

    try:
        python_type = column.type.python_type
    except NotImplementedError:
        return "text"

    if python_type is bool:
        return "boolean"
    if python_type is int:
        return "number"
    if python_type is float:
        return "number"
    if python_type is Decimal:
        return "decimal"
    if python_type is datetime:
        return "datetime"
    if python_type is UUID:
        return "text"
    if str(column.type).lower() == "date":
        return "date"
    if str(column.type).lower().startswith("varchar") and "password" in field_name.lower():
        return "password"
    return "text"


def _get_configured_field_names(config: AdminModelConfig) -> list[str]:
    names: list[str] = []
    for field_group in (
            config.list_display,
            config.list_fields,
            config.detail_fields,
            config.create_fields,
            config.update_fields,
    ):
        if field_group is None:
            continue
        for field_name in field_group:
            if field_name not in names:
                names.append(field_name)
    return names


def _serialize_related_pk(instance: Any) -> Any:
    mapper = sa_inspect(type(instance))
    pk_column = mapper.primary_key[0]
    return getattr(instance, pk_column.key)


def _get_list_filter_input_kind(config: AdminModelConfig, list_filter: Any) -> str:
    if list_filter.input_kind is not None:
        return list_filter.input_kind

    if list_filter.options:
        return "select"

    if list_filter.relation_model is not None:
        return "select"

    if list_filter.field_name is None:
        return "text"

    mapper = sa_inspect(config.model)
    if list_filter.field_name in mapper.relationships:
        return "select"
    if list_filter.field_name not in mapper.columns:
        return "text"

    column = mapper.columns[list_filter.field_name]
    try:
        python_type = column.type.python_type
    except NotImplementedError:
        return "text"

    if python_type is bool:
        return "boolean"
    return "text"


def _list_filter_has_choices(config: AdminModelConfig, list_filter: Any) -> bool:
    if list_filter.options:
        return True

    input_kind = _get_list_filter_input_kind(config, list_filter)
    if input_kind == "boolean":
        return True
    if list_filter.relation_model is not None:
        return True
    if list_filter.field_name is None:
        return False

    mapper = sa_inspect(config.model)
    return list_filter.field_name in mapper.relationships


def _get_list_filter_options(locale: str, config: AdminModelConfig, list_filter: Any) -> list[dict[str, str]]:
    input_kind = _get_list_filter_input_kind(config, list_filter)
    if list_filter.options:
        return [{"value": option.value, "label": option.label} for option in list_filter.options]
    if input_kind == "boolean":
        return [
            {"value": "true", "label": translate(locale, "yes")},
            {"value": "false", "label": translate(locale, "no")},
        ]
    return []



from __future__ import annotations

import inspect
from typing import Any

from fastapi import APIRouter, Depends, HTTPException, Query, Request, Response, status
from sqlalchemy import func, or_, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.inspection import inspect as sa_inspect
from sqlalchemy.orm import selectinload

from xladmin.config import HttpConfig, ModelConfig
from xladmin.delete_preview import build_delete_plan, build_delete_preview
from xladmin.i18n import translate
from xladmin.introspection import (
    convert_value_for_column,
    get_column_names,
    get_create_fields,
    get_model_blocks_meta,
    get_model_meta,
    get_pk_field_name,
    get_sort_column_name,
    get_update_fields,
)
from xladmin.registry import build_registry
from xladmin.serializer import serialize_model_instance


def create_router(config: HttpConfig) -> APIRouter:
    registry = build_registry(config.registry)
    router = APIRouter(prefix="/xladmin", tags=["XLAdmin"])
    model_meta_by_slug = {
        model_config.slug: get_model_meta(model_config, locale=registry.locale)
        for model_config in registry.list()
    }
    models_response = {
        "locale": registry.locale,
        "items": list(model_meta_by_slug.values()),
        "blocks": get_model_blocks_meta(registry),
    }

    def _check_access(user: Any) -> None:
        if not config.is_allowed(user):
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail=translate(registry.locale, "forbidden"))

    async def _get_model_config(slug: str) -> ModelConfig:
        try:
            return registry.get(slug)
        except KeyError as exc:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=translate(registry.locale, "admin_model_not_found"),
            ) from exc

    @router.get("/models/")
    async def list_models(user: Any = Depends(config.get_current_user_dependency)) -> dict[str, Any]:
        _check_access(user)
        return models_response

    @router.get("/models/{slug}/")
    async def get_model(
            slug: str,
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        _check_access(user)
        try:
            return model_meta_by_slug[slug]
        except KeyError:
            await _get_model_config(slug)
            raise

    @router.get("/models/{slug}/items/")
    async def list_items(
            slug: str,
            request: Request,
            limit: int = Query(default=50, ge=1, le=500),
            offset: int = Query(default=0, ge=0),
            q: str | None = None,
            sort: str | None = None,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        _check_access(user)
        model_config = await _get_model_config(slug)
        request_query_params = request.query_params if request is not None else {}
        list_filter_values: dict[str, str] = {}
        for list_filter in model_config.list_filters:
            raw_value = request_query_params.get(list_filter.slug)
            if raw_value in (None, ""):
                continue
            list_filter_values[list_filter.slug] = str(raw_value)
        query = select(model_config.model)
        if model_config.query_for_list is not None:
            custom_query = model_config.query_for_list(query, session, user)
            query = await custom_query if inspect.isawaitable(custom_query) else custom_query
        query = _apply_eager_loads(model_config, query)
        query = _apply_search(model_config, query, q, session)
        query = await _apply_list_filters(model_config, query, list_filter_values, session, user)
        query = _apply_ordering(model_config, query, sort)
        total_query = select(func.count()).select_from(query.subquery())
        total = int((await session.execute(total_query)).scalar_one())
        items = list((await session.execute(query.limit(limit).offset(offset))).scalars())
        return {
            "meta": model_meta_by_slug[model_config.slug],
            "pagination": {"limit": limit, "offset": offset, "total": total},
            "items": [serialize_model_instance(model_config, item, mode="list") for item in items],
        }

    @router.get("/models/{slug}/items/{item_id}/")
    async def get_item(
            slug: str,
            item_id: str,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        _check_access(user)
        model_config = await _get_model_config(slug)
        item = await _get_item_by_pk(session, model_config, item_id)
        if item is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=translate(registry.locale, "object_not_found"),
            )
        return {
            "meta": model_meta_by_slug[model_config.slug],
            "item": serialize_model_instance(model_config, item, mode="detail"),
        }

    @router.post("/models/{slug}/items/", status_code=status.HTTP_201_CREATED)
    async def create_item(
            slug: str,
            payload: dict[str, Any],
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        _check_access(user)
        model_config = await _get_model_config(slug)
        item = model_config.model()
        await _apply_payload_to_item(session, model_config, item, payload, mode="create")
        session.add(item)
        await session.commit()
        await session.refresh(item)
        return {"item": serialize_model_instance(model_config, item, mode="detail")}

    @router.patch("/models/{slug}/items/{item_id}/")
    async def patch_item(
            slug: str,
            item_id: str,
            payload: dict[str, Any],
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        _check_access(user)
        model_config = await _get_model_config(slug)
        item = await _get_item_by_pk(session, model_config, item_id)
        if item is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=translate(registry.locale, "object_not_found"),
            )
        await _apply_payload_to_item(session, model_config, item, payload, mode="update")
        await session.commit()
        await session.refresh(item)
        return {"item": serialize_model_instance(model_config, item, mode="detail")}

    @router.get("/models/{slug}/items/{item_id}/delete-preview/")
    async def get_delete_preview(
            slug: str,
            item_id: str,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        _check_access(user)
        model_config = await _get_model_config(slug)
        item = await _get_item_by_pk(session, model_config, item_id)
        if item is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=translate(registry.locale, "object_not_found"),
            )
        return await build_delete_preview(session, registry, model_config, [item])

    @router.delete("/models/{slug}/items/{item_id}/", status_code=status.HTTP_204_NO_CONTENT)
    async def delete_item(
            slug: str,
            item_id: str,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> Response:
        _check_access(user)
        model_config = await _get_model_config(slug)
        item = await _get_item_by_pk(session, model_config, item_id)
        if item is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=translate(registry.locale, "object_not_found"),
            )
        preview, delete_items, set_null_items = await build_delete_plan(session, registry, model_config, [item])
        if not preview["can_delete"]:
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail=translate(registry.locale, "delete_blocked"),
            )
        for related_item, attribute_names in set_null_items:
            for attribute_name in attribute_names:
                setattr(related_item, attribute_name, None)
        for related_delete_item in delete_items:
            await session.delete(related_delete_item)
        await session.commit()
        return Response(status_code=status.HTTP_204_NO_CONTENT)

    @router.post("/models/{slug}/items/{item_id}/actions/{action_slug}/")
    async def object_action(
            slug: str,
            item_id: str,
            action_slug: str,
            payload: dict[str, Any] | None = None,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        _check_access(user)
        model_config = await _get_model_config(slug)
        item = await _get_item_by_pk(session, model_config, item_id)
        if item is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=translate(registry.locale, "object_not_found"),
            )

        action = next((item for item in model_config.object_actions if item.slug == action_slug), None)
        if action is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=translate(registry.locale, "object_action_not_found"),
            )

        result = action.handler(session, model_config, item, payload or {}, user)
        if inspect.isawaitable(result):
            result = await result
        await session.commit()
        await session.refresh(item)
        item = await _get_item_by_pk(session, model_config, item_id)
        return {
            "item": serialize_model_instance(model_config, item, mode="detail"),
            "result": result or {},
        }

    @router.post("/models/{slug}/bulk-delete/")
    async def bulk_delete(
            slug: str,
            payload: dict[str, list[Any]],
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, int]:
        _check_access(user)
        model_config = await _get_model_config(slug)
        ids = payload.get("ids", [])
        if not ids:
            return {"deleted": 0}
        pk_attr = getattr(model_config.model, get_pk_field_name(model_config))
        query = select(model_config.model).where(pk_attr.in_([_convert_pk(item_id) for item_id in ids]))
        items = list((await session.execute(query)).scalars())
        preview, delete_items, set_null_items = await build_delete_plan(session, registry, model_config, items)
        if not preview["can_delete"]:
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail=translate(registry.locale, "delete_blocked"),
            )
        for related_item, attribute_names in set_null_items:
            for attribute_name in attribute_names:
                setattr(related_item, attribute_name, None)
        for item in delete_items:
            await session.delete(item)
        await session.commit()
        return {"deleted": len(items)}

    @router.post("/models/{slug}/bulk-delete-preview/")
    async def bulk_delete_preview(
            slug: str,
            payload: dict[str, list[Any]],
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        _check_access(user)
        model_config = await _get_model_config(slug)
        ids = payload.get("ids", [])
        if not ids:
            return {
                "can_delete": True,
                "summary": {"roots": 0, "delete": 0, "protect": 0, "set_null": 0, "total": 0},
                "roots": [],
            }
        pk_attr = getattr(model_config.model, get_pk_field_name(model_config))
        query = select(model_config.model).where(pk_attr.in_([_convert_pk(item_id) for item_id in ids]))
        items = list((await session.execute(_apply_eager_loads(model_config, query))).scalars())
        return await build_delete_preview(session, registry, model_config, items)

    @router.post("/models/{slug}/bulk-actions/{action_slug}/")
    async def bulk_action(
            slug: str,
            action_slug: str,
            payload: dict[str, Any],
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        _check_access(user)
        model_config = await _get_model_config(slug)
        ids = payload.get("ids", [])
        if not ids:
            return {"processed": 0}
        if action_slug == "delete":
            deleted_response = await bulk_delete(slug, {"ids": ids}, session, user)
            return {"processed": deleted_response["deleted"]}

        action = next((item for item in model_config.bulk_actions if item.slug == action_slug), None)
        if action is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=translate(registry.locale, "bulk_action_not_found"),
            )

        pk_attr = getattr(model_config.model, get_pk_field_name(model_config))
        query = select(model_config.model).where(pk_attr.in_([_convert_pk(item_id) for item_id in ids]))
        items = list((await session.execute(_apply_eager_loads(model_config, query))).scalars())
        result = action.handler(session, model_config, items, payload, user)
        if inspect.isawaitable(result):
            result = await result
        await session.commit()
        return {"processed": len(items), **(result or {})}

    @router.get("/models/{slug}/fields/{field_name}/choices/")
    async def field_choices(
            slug: str,
            field_name: str,
            q: str | None = None,
            ids: str | None = None,
            limit: int = Query(default=25, ge=1, le=100),
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        _check_access(user)
        model_config = await _get_model_config(slug)
        relation_model = _resolve_relation_model(model_config, field_name, locale=registry.locale)
        relation_config = registry.find_by_model(relation_model)
        label_field = (
                model_config.get_field_config(field_name).relation_label_field
                or _resolve_label_field(relation_model)
        )
        query = select(relation_model)
        relation_pk_name = sa_inspect(relation_model).primary_key[0].key
        selected_ids = [_convert_pk(item_id) for item_id in ids.split(",") if item_id] if ids else []

        if selected_ids:
            query = query.where(getattr(relation_model, relation_pk_name).in_(selected_ids))

        if q:
            if relation_config is not None:
                query = _apply_search(relation_config, query, q, session)
            elif hasattr(relation_model, label_field):
                query = query.where(getattr(relation_model, label_field).ilike(f"%{q}%"))

        items = list((await session.execute(query.limit(limit))).scalars())
        return {
            "items": [
                {
                    "id": getattr(item, relation_pk_name),
                    "label": str(getattr(item, label_field, None) or getattr(item, relation_pk_name)),
                }
                for item in items
            ],
        }

    @router.get("/models/{slug}/filters/{filter_slug}/choices/")
    async def filter_choices(
            slug: str,
            filter_slug: str,
            q: str | None = None,
            ids: str | None = None,
            limit: int = Query(default=25, ge=1, le=100),
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        _check_access(user)
        model_config = await _get_model_config(slug)
        list_filter = next((item for item in model_config.list_filters if item.slug == filter_slug), None)
        if list_filter is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=translate(registry.locale, "list_filter_not_found"),
            )

        input_kind = list_filter.input_kind or _resolve_list_filter_input_kind_for_meta(model_config, list_filter)
        if input_kind == "boolean":
            return {
                "items": [
                    {"id": "true", "label": translate(registry.locale, "yes")},
                    {"id": "false", "label": translate(registry.locale, "no")},
                ],
            }
        if list_filter.options:
            return {
                "items": [
                    {"id": option.value, "label": option.label}
                    for option in list_filter.options
                ],
            }

        relation_model = _resolve_list_filter_relation_model(model_config, list_filter, locale=registry.locale)
        if relation_model is None:
            return {"items": []}

        relation_config = registry.find_by_model(relation_model)
        label_field = list_filter.relation_label_field or _resolve_label_field(relation_model)
        query = select(relation_model)
        relation_pk_name = sa_inspect(relation_model).primary_key[0].key
        selected_ids = [_convert_pk(item_id) for item_id in ids.split(",") if item_id] if ids else []

        if relation_config is not None and relation_config.query_for_list is not None:
            custom_query = relation_config.query_for_list(query, session, user)
            query = await custom_query if inspect.isawaitable(custom_query) else custom_query

        if selected_ids:
            query = query.where(getattr(relation_model, relation_pk_name).in_(selected_ids))

        if q:
            if relation_config is not None:
                query = _apply_search(relation_config, query, q, session)
            elif hasattr(relation_model, label_field):
                query = query.where(getattr(relation_model, label_field).ilike(f"%{q}%"))

        if hasattr(relation_model, label_field):
            query = query.order_by(getattr(relation_model, label_field).asc())

        items = list((await session.execute(query.limit(limit))).scalars())
        return {
            "items": [
                {
                    "id": getattr(item, relation_pk_name),
                    "label": str(getattr(item, label_field, None) or getattr(item, relation_pk_name)),
                }
                for item in items
            ],
        }

    return router


async def _apply_payload_to_item(
        session: AsyncSession,
        model_config: ModelConfig,
        item: Any,
        payload: dict[str, Any],
        *,
        mode: str,
) -> None:
    mapper = sa_inspect(model_config.model)
    column_names = set(get_column_names(model_config))
    relationship_names = set(mapper.relationships.keys())
    allowed_fields = set(get_create_fields(model_config) if mode == "create" else get_update_fields(model_config))
    for field_name, raw_value in payload.items():
        if field_name not in allowed_fields:
            continue
        field_config = model_config.get_field_config(field_name)
        value = field_config.value_parser(raw_value) if field_config.value_parser is not None else raw_value

        if field_config.value_setter is not None:
            field_config.value_setter(item, value, payload, mode)
            continue

        if field_name in relationship_names:
            await _assign_relationship_value(session, model_config, item, field_name, value)
            continue

        if field_name not in column_names:
            continue

        column = mapper.columns[field_name]
        setattr(item, field_name, convert_value_for_column(column, value))


def _apply_ordering(model_config: ModelConfig, query, sort: str | None = None):
    ordering_items = (
        [item.strip() for item in (sort or "").split(",") if item.strip()]
        if sort
        else list(model_config.ordering)
    )
    if not ordering_items:
        return query

    for item in ordering_items:
        descending = item.startswith("-")
        field_name = item[1:] if descending else item
        sort_column_name = get_sort_column_name(model_config, field_name)
        if sort_column_name is None:
            continue
        column = getattr(model_config.model, sort_column_name)
        query = query.order_by(column.desc() if descending else column.asc())
    return query


def _apply_search(model_config: ModelConfig, query, q: str | None, session: AsyncSession):
    if not q:
        return query
    if model_config.search_query_builder is not None:
        return model_config.search_query_builder(query, q, session)
    search_fields = model_config.search_fields or tuple(
        column_name
        for column_name, column in sa_inspect(model_config.model).columns.items()
        if getattr(column.type, "python_type", None) is str
    )
    conditions = [getattr(model_config.model, field_name).ilike(f"%{q}%") for field_name in search_fields]
    if not conditions:
        return query
    return query.where(or_(*conditions))


async def _apply_list_filters(
        model_config: ModelConfig,
        query,
        values: dict[str, str],
        session: AsyncSession,
        user: Any,
):
    if not values:
        return query

    mapper = sa_inspect(model_config.model)
    for list_filter in model_config.list_filters:
        raw_value = values.get(list_filter.slug)
        if raw_value in (None, ""):
            continue

        filter_value = str(raw_value)
        parsed_value = list_filter.value_parser(filter_value) if list_filter.value_parser is not None else filter_value
        option_config = next((item for item in list_filter.options if item.value == filter_value), None)
        filter_handler = option_config.filter_handler if option_config is not None else list_filter.filter_handler
        if filter_handler is not None:
            custom_query = filter_handler(query, parsed_value, session, user)
            query = await custom_query if inspect.isawaitable(custom_query) else custom_query
            continue

        if list_filter.field_name is None:
            continue

        if list_filter.field_name in mapper.relationships:
            relationship = mapper.relationships[list_filter.field_name]
            related_pk_name = relationship.mapper.primary_key[0].key
            related_pk_attr = getattr(relationship.mapper.class_, related_pk_name)
            if relationship.uselist:
                query = query.where(
                    getattr(model_config.model, list_filter.field_name).any(
                        related_pk_attr == _convert_pk(parsed_value),
                    ),
                )
                continue
            local_columns = list(relationship.local_columns)
            if not local_columns:
                continue
            query = query.where(getattr(model_config.model, local_columns[0].key) == _convert_pk(parsed_value))
            continue

        if list_filter.field_name not in mapper.columns:
            continue

        column = mapper.columns[list_filter.field_name]
        column_attr = getattr(model_config.model, list_filter.field_name)
        input_kind = list_filter.input_kind or _resolve_list_filter_input_kind(column, list_filter)
        if input_kind == "boolean":
            query = query.where(column_attr.is_(_parse_boolean_value(parsed_value)))
            continue
        if input_kind == "text":
            query = query.where(column_attr.ilike(f"%{parsed_value}%"))
            continue
        query = query.where(column_attr == convert_value_for_column(column, parsed_value))

    return query


def _apply_eager_loads(model_config: ModelConfig, query):
    mapper = sa_inspect(model_config.model)
    relation_names = set(mapper.relationships.keys())
    relation_fields = [
        field_name
        for field_name in (
            *(model_config.list_display or ()),
            *(model_config.detail_fields or ()),
            *model_config.fields.keys(),
        )
        if field_name in relation_names
    ]
    for field_name in dict.fromkeys(relation_fields):
        query = query.options(selectinload(getattr(model_config.model, field_name)))
    return query


def _resolve_list_filter_input_kind(column: Any, list_filter: Any) -> str:
    if list_filter.options:
        return "select"
    try:
        python_type = column.type.python_type
    except NotImplementedError:
        return "text"
    if python_type is bool:
        return "boolean"
    return "text"


def _resolve_list_filter_input_kind_for_meta(model_config: ModelConfig, list_filter: Any) -> str:
    if list_filter.input_kind is not None:
        return list_filter.input_kind
    if list_filter.options or list_filter.relation_model is not None:
        return "select"
    if list_filter.field_name is None:
        return "text"

    mapper = sa_inspect(model_config.model)
    if list_filter.field_name in mapper.relationships:
        return "select"
    if list_filter.field_name not in mapper.columns:
        return "text"
    return _resolve_list_filter_input_kind(mapper.columns[list_filter.field_name], list_filter)


def _parse_boolean_value(value: Any) -> bool:
    if isinstance(value, bool):
        return value
    normalized_value = str(value).strip().lower()
    return normalized_value in {"1", "true", "yes", "y", "on"}


def _resolve_relation_model(model_config: ModelConfig, field_name: str, *, locale: str | None = None) -> type[Any]:
    configured_model = model_config.get_field_config(field_name).relation_model
    if configured_model is not None:
        return configured_model

    mapper = sa_inspect(model_config.model)
    if field_name in mapper.relationships:
        return mapper.relationships[field_name].mapper.class_
    column = mapper.columns[field_name]
    foreign_key = next(iter(column.foreign_keys), None)
    if foreign_key is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=translate(locale, "field_has_no_relation_choices"),
        )
    remote_table = foreign_key.column.table
    for registry_mapper in mapper.registry.mappers:
        model = registry_mapper.class_
        if getattr(model, "__table__", None) is remote_table:
            return model
    raise HTTPException(
        status_code=status.HTTP_400_BAD_REQUEST,
        detail=translate(locale, "related_model_not_found"),
    )


def _resolve_list_filter_relation_model(
        model_config: ModelConfig,
        list_filter: Any,
        *,
        locale: str | None = None,
) -> type[Any] | None:
    if list_filter.relation_model is not None:
        return list_filter.relation_model
    if list_filter.field_name is None:
        return None

    mapper = sa_inspect(model_config.model)
    if list_filter.field_name in mapper.relationships:
        return mapper.relationships[list_filter.field_name].mapper.class_
    if list_filter.field_name in mapper.columns and mapper.columns[list_filter.field_name].foreign_keys:
        return _resolve_relation_model(model_config, list_filter.field_name, locale=locale)
    return None


def _resolve_label_field(model: type[Any]) -> str:
    for candidate in ("name", "title", "username", "slug", "email", "code_name"):
        if hasattr(model, candidate):
            return candidate
    return sa_inspect(model).primary_key[0].key


def _convert_pk(value: Any) -> Any:
    if isinstance(value, int):
        return value
    if isinstance(value, str) and value.isdigit():
        return int(value)
    return value


async def _assign_relationship_value(
        session: AsyncSession,
        model_config: ModelConfig,
        item: Any,
        field_name: str,
        value: Any,
) -> None:
    mapper = sa_inspect(model_config.model)
    relationship = mapper.relationships[field_name]
    relation_model = _resolve_relation_model(model_config, field_name)
    relation_pk_name = sa_inspect(relation_model).primary_key[0].key

    if relationship.uselist:
        raw_ids = list(value or [])
        if not raw_ids:
            setattr(item, field_name, [])
            return
        normalized_ids = [_convert_pk(str(item_id)) for item_id in raw_ids]
        query = select(relation_model).where(getattr(relation_model, relation_pk_name).in_(normalized_ids))
        related_items = list((await session.execute(query)).scalars())
        setattr(item, field_name, related_items)
        return

    if value in (None, ""):
        setattr(item, field_name, None)
        return

    related_item = await session.get(relation_model, _convert_pk(str(value)))
    setattr(item, field_name, related_item)


async def _get_item_by_pk(
        session: AsyncSession,
        model_config: ModelConfig,
        item_id: str,
) -> Any:
    pk_attr = getattr(model_config.model, get_pk_field_name(model_config))
    query = select(model_config.model).where(pk_attr == _convert_pk(item_id))
    query = _apply_eager_loads(model_config, query)
    return (await session.execute(query)).scalar_one_or_none()


create_admin_router = create_router



from __future__ import annotations

from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from typing import Any, Literal

FieldInputKind = Literal[
    "text",
    "textarea",
    "password",
    "number",
    "decimal",
    "boolean",
    "date",
    "datetime",
    "json",
    "relation",
    "relation-multiple",
]

FieldDisplayKind = Literal["text", "image"]

ListFilterInputKind = Literal["text", "select", "boolean"]
AdminLocale = Literal["ru", "en"]


@dataclass(slots=True)
class FieldConfig:
    label: str | None = None
    help_text: str | None = None
    width_px: int | None = None
    display_kind: FieldDisplayKind | None = None
    image_url_prefix: str | None = None
    hidden_in_list: bool = False
    hidden_in_detail: bool = False
    hidden_in_form: bool = False
    read_only: bool = False
    input_kind: FieldInputKind | None = None
    ordering_field: str | None = None
    value_getter: Callable[[Any], Any] | None = None
    value_parser: Callable[[Any], Any] | None = None
    value_setter: Callable[[Any, Any, dict[str, Any], str], None] | None = None
    relation_label_field: str | None = None
    relation_model: type[Any] | None = None


@dataclass(slots=True)
class ListFilterOptionConfig:
    value: str
    label: str
    filter_handler: Callable[[Any, Any, Any, Any], Awaitable[Any] | Any] | None = None


@dataclass(slots=True)
class ListFilterConfig:
    slug: str
    label: str
    group: str | None = None
    field_name: str | None = None
    input_kind: ListFilterInputKind | None = None
    placeholder: str | None = None
    options: tuple[ListFilterOptionConfig, ...] = ()
    value_parser: Callable[[str], Any] | None = None
    filter_handler: Callable[[Any, Any, Any, Any], Awaitable[Any] | Any] | None = None
    relation_label_field: str | None = None
    relation_model: type[Any] | None = None


@dataclass(slots=True)
class BulkActionConfig:
    slug: str
    label: str
    handler: Callable[
        [Any, ModelConfig, list[Any], dict[str, Any], Any],
        Awaitable[dict[str, Any] | None] | dict[str, Any] | None,
    ]


@dataclass(slots=True)
class ObjectActionConfig:
    slug: str
    label: str
    handler: Callable[
        [Any, ModelConfig, Any, dict[str, Any], Any],
        Awaitable[dict[str, Any] | None] | dict[str, Any] | None,
    ]


@dataclass(slots=True)
class ModelsBlockConfig:
    slug: str
    title: str
    models: tuple[type[Any] | str, ...] = ()
    model_slugs: tuple[str, ...] = ()
    description: str | None = None
    color: str | None = None
    collapsible: bool = False
    default_expanded: bool = True


@dataclass(slots=True)
class ModelConfig:
    model: type[Any]
    slug: str | None = None
    title: str | None = None
    description: str | None = None
    fields: dict[str, FieldConfig] = field(default_factory=dict)
    pk_field: str = "id"
    display_field: str | None = None
    search_fields: tuple[str, ...] = ()
    search_query_builder: Callable[[Any, str, Any], Any] | None = None
    query_for_list: Callable[[Any, Any, Any], Awaitable[Any] | Any] | None = None
    list_filters: tuple[ListFilterConfig, ...] = ()
    list_display: tuple[str, ...] | None = None
    list_fields: tuple[str, ...] | None = None
    detail_fields: tuple[str, ...] | None = None
    create_fields: tuple[str, ...] | None = None
    update_fields: tuple[str, ...] | None = None
    ordering: tuple[str, ...] = ()
    page_size: int = 50
    bulk_actions: tuple[BulkActionConfig, ...] = ()
    object_actions: tuple[ObjectActionConfig, ...] = ()

    def get_field_config(self, field_name: str) -> FieldConfig:
        return self.fields.get(field_name, FieldConfig())


@dataclass(slots=True)
class AdminConfig:
    models: tuple[ModelConfig, ...] = ()
    models_blocks: tuple[ModelsBlockConfig, ...] = ()
    locale: AdminLocale = "ru"


@dataclass(slots=True)
class HttpConfig:
    registry: Any
    get_db_session_dependency: Callable[..., Any]
    get_current_user_dependency: Callable[..., Any]
    is_allowed: Callable[[Any], bool]


AdminFieldConfig = FieldConfig
AdminListFilterConfig = ListFilterConfig
AdminListFilterOptionConfig = ListFilterOptionConfig
AdminBulkActionConfig = BulkActionConfig
AdminObjectActionConfig = ObjectActionConfig
AdminModelsBlockConfig = ModelsBlockConfig
AdminModelConfig = ModelConfig
AdminHTTPConfig = HttpConfig
ModelsBlock = ModelsBlockConfig



from __future__ import annotations

from collections.abc import AsyncIterator
from datetime import date
from types import SimpleNamespace
from typing import Any

import pytest
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient
from sqlalchemy import Date, ForeignKey, String, select
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship
from sqlalchemy.pool import StaticPool

from xladmin import (
    AdminBulkActionConfig,
    AdminConfig,
    AdminFieldConfig,
    AdminHTTPConfig,
    AdminListFilterConfig,
    AdminListFilterOptionConfig,
    AdminModelConfig,
    AdminObjectActionConfig,
    ModelConfig,
    create_admin_router,
)


# TODO: разбить огромный файл на мелкие
class Base(DeclarativeBase):
    pass


class DemoUserORM(Base):
    __tablename__ = "demo_users"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    username: Mapped[str] = mapped_column(String(64), nullable=False, unique=True)
    password: Mapped[str] = mapped_column(String(255), nullable=False, default="")
    joined_on: Mapped[date | None] = mapped_column(Date)
    is_active: Mapped[bool] = mapped_column(default=True)
    posts: Mapped[list[DemoPostORM]] = relationship(
        back_populates="user",
        cascade="all, delete-orphan",
    )
    badges: Mapped[list[DemoBadgeORM]] = relationship(back_populates="user")
    secrets: Mapped[list[DemoSecretORM]] = relationship(
        back_populates="user",
        cascade="all, delete-orphan",
    )
    tasks: Mapped[list[DemoTaskORM]] = relationship(back_populates="user")
    roles: Mapped[list[DemoRoleORM]] = relationship(
        secondary="demo_user_roles",
        back_populates="users",
    )


class DemoRoleORM(Base):
    __tablename__ = "demo_roles"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(String(64), nullable=False, unique=True)
    users: Mapped[list[DemoUserORM]] = relationship(
        secondary="demo_user_roles",
        back_populates="roles",
    )


class DemoUserRoleORM(Base):
    __tablename__ = "demo_user_roles"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    user_id: Mapped[int] = mapped_column(ForeignKey("demo_users.id"), nullable=False)
    role_id: Mapped[int] = mapped_column(ForeignKey("demo_roles.id"), nullable=False)


class DemoAuditORM(Base):
    __tablename__ = "demo_audits"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    role_id: Mapped[int | None] = mapped_column(ForeignKey("demo_roles.id"), nullable=True)


class DemoPostORM(Base):
    __tablename__ = "demo_posts"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    title: Mapped[str] = mapped_column(String(128), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("demo_users.id"), nullable=False)
    user: Mapped[DemoUserORM] = relationship(back_populates="posts")
    comments: Mapped[list[DemoCommentORM]] = relationship(
        back_populates="post",
        cascade="all, delete-orphan",
    )


class DemoCommentORM(Base):
    __tablename__ = "demo_comments"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    body: Mapped[str] = mapped_column(String(255), nullable=False)
    post_id: Mapped[int] = mapped_column(ForeignKey("demo_posts.id"), nullable=False)
    post: Mapped[DemoPostORM] = relationship(back_populates="comments")


class DemoBadgeORM(Base):
    __tablename__ = "demo_badges"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    title: Mapped[str] = mapped_column(String(64), nullable=False)
    user_id: Mapped[int | None] = mapped_column(ForeignKey("demo_users.id"), nullable=True)
    user: Mapped[DemoUserORM | None] = relationship(back_populates="badges")


class DemoSecretORM(Base):
    __tablename__ = "demo_secrets"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    code_name: Mapped[str] = mapped_column(String(64), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("demo_users.id"), nullable=False)
    user: Mapped[DemoUserORM] = relationship(back_populates="secrets")


class DemoTaskORM(Base):
    __tablename__ = "demo_tasks"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    title: Mapped[str] = mapped_column(String(64), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("demo_users.id"), nullable=False)
    user: Mapped[DemoUserORM] = relationship(back_populates="tasks")


class DemoOwnedProjectORM(Base):
    __tablename__ = "demo_owned_projects"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    title: Mapped[str] = mapped_column(String(64), nullable=False)
    author_id: Mapped[int] = mapped_column(ForeignKey("demo_users.id"), nullable=False)


class DemoOwnedBoardORM(Base):
    __tablename__ = "demo_owned_boards"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    title: Mapped[str] = mapped_column(String(64), nullable=False)
    project_id: Mapped[int] = mapped_column(ForeignKey("demo_owned_projects.id"), nullable=False)


class DemoTokenProfileORM(Base):
    __tablename__ = "demo_token_profiles"

    tokenprofile_ptr_id: Mapped[str] = mapped_column(String(36), primary_key=True)
    title: Mapped[str] = mapped_column(String(64), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("demo_users.id"), nullable=False)


class DemoAliasedMessageORM(Base):
    __tablename__ = "demo_aliased_messages"

    id: Mapped[str] = mapped_column(String(36), primary_key=True)


class DemoAliasedChatORM(Base):
    __tablename__ = "demo_aliased_chats"

    id: Mapped[str] = mapped_column(String(36), primary_key=True)
    last_branch_message_id: Mapped[str | None] = mapped_column(
        "last_leaf_message_id",
        String(36),
        ForeignKey("demo_aliased_messages.id"),
    )


class DemoSoftUserORM(Base):
    __tablename__ = "demo_soft_users"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    username: Mapped[str] = mapped_column(String(64), nullable=False)


class DemoSoftNoteORM(Base):
    __tablename__ = "demo_soft_notes"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    body: Mapped[str] = mapped_column(String(255), nullable=False)
    user_id: Mapped[int | None] = mapped_column(ForeignKey("demo_soft_users.id", ondelete="SET NULL"), nullable=True)


class DemoHardParentORM(Base):
    __tablename__ = "demo_hard_parents"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    title: Mapped[str] = mapped_column(String(64), nullable=False)


class DemoHardChildORM(Base):
    __tablename__ = "demo_hard_children"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    title: Mapped[str] = mapped_column(String(64), nullable=False)
    parent_id: Mapped[int] = mapped_column(ForeignKey("demo_hard_parents.id", ondelete="CASCADE"), nullable=False)


async def _build_app(locale: str = "ru", *, is_staff: bool = True) -> tuple[FastAPI, async_sessionmaker[AsyncSession]]:
    engine = create_async_engine(
        "sqlite+aiosqlite:///:memory:",
        poolclass=StaticPool,
    )
    session_factory = async_sessionmaker(engine, expire_on_commit=False)

    async def get_session() -> AsyncIterator[AsyncSession]:
        async with session_factory() as session:
            yield session

    def display_name(instance: DemoUserORM) -> str:
        return f"Пользователь {instance.id}: {instance.username}"

    def hash_password_setter(
            instance: DemoUserORM,
            value: str,
            _payload: dict[str, object],
            _mode: str,
    ) -> None:
        instance.password = f"hashed::{value}"

    def custom_user_search(query, q: str, _session: AsyncSession):
        return query.where(DemoUserORM.username.ilike(f"{q}%"))

    def custom_role_search(query, q: str, _session: AsyncSession):
        return query.where(DemoRoleORM.name.ilike(f"{q}%"))

    def filter_roles_for_list(query, _session: AsyncSession, _user: Any):
        return query.where(DemoRoleORM.name != "hidden")

    def filter_users_by_prefix(query, value: str, _session: AsyncSession, _user: Any):
        if not value:
            return query
        return query.where(DemoUserORM.username.ilike(f"{value}%"))

    def filter_active_audience(query, _value: str, _session: AsyncSession, _user: Any):
        return query.where(DemoUserORM.is_active.is_(True))

    def filter_inactive_audience(query, _value: str, _session: AsyncSession, _user: Any):
        return query.where(DemoUserORM.is_active.is_(False))

    async def activate_users(
            _session: AsyncSession,
            _model_config: AdminModelConfig,
            items: list[DemoUserORM],
            _payload: dict[str, Any],
            _user: Any,
    ) -> dict[str, Any]:
        for item in items:
            item.is_active = True
        return {"activated": len(items)}

    async def deactivate_user(
            _session: AsyncSession,
            _model_config: AdminModelConfig,
            item: DemoUserORM,
            _payload: dict[str, Any],
            _user: Any,
    ) -> dict[str, Any]:
        item.is_active = False
        return {"deactivated": 1}

    config = AdminConfig(
        locale=locale,
        models=(
            ModelConfig(
                model=DemoUserORM,
                slug="users",
                title="Пользователи",
                list_display=("id", "display_name", "joined_on", "is_active"),
                detail_fields=("id", "username", "joined_on", "is_active", "new_password"),
                create_fields=("username", "joined_on", "is_active", "new_password"),
                update_fields=("username", "joined_on", "is_active", "new_password"),
                search_query_builder=custom_user_search,
                page_size=120,
                ordering=("-id",),
                list_filters=(
                    AdminListFilterConfig(
                        slug="status",
                        label="Status",
                        group="Flags",
                        field_name="is_active",
                        input_kind="boolean",
                    ),
                    AdminListFilterConfig(
                        slug="role_id",
                        label="Role",
                        group="Access",
                        field_name="roles",
                        relation_model=DemoRoleORM,
                        relation_label_field="name",
                    ),
                    AdminListFilterConfig(
                        slug="username_contains",
                        label="Username",
                        field_name="username",
                        input_kind="text",
                    ),
                    AdminListFilterConfig(
                        slug="username_prefix",
                        label="Username Prefix",
                        filter_handler=filter_users_by_prefix,
                    ),
                    AdminListFilterConfig(
                        slug="audience",
                        label="Audience",
                        group="Flags",
                        options=(
                            AdminListFilterOptionConfig(
                                value="active-only",
                                label="Active only",
                                filter_handler=filter_active_audience,
                            ),
                            AdminListFilterOptionConfig(
                                value="inactive-only",
                                label="Inactive only",
                                filter_handler=filter_inactive_audience,
                            ),
                        ),
                    ),
                ),
                fields={
                    "display_name": AdminFieldConfig(
                        label="Отображение",
                        width_px=320,
                        read_only=True,
                        hidden_in_form=True,
                        ordering_field="username",
                        value_getter=display_name,
                    ),
                    "new_password": AdminFieldConfig(
                        label="Новый пароль",
                        input_kind="password",
                        value_getter=lambda instance: "",
                        value_setter=hash_password_setter,
                    ),
                },
                bulk_actions=(
                    AdminBulkActionConfig(
                        slug="activate",
                        label="Активировать",
                        handler=activate_users,
                    ),
                ),
                object_actions=(
                    AdminObjectActionConfig(
                        slug="deactivate",
                        label="Деактивировать",
                        handler=deactivate_user,
                    ),
                ),
            ),
            ModelConfig(
                model=DemoRoleORM,
                slug="roles",
                title="Роли",
                list_display=("id", "name"),
                search_query_builder=custom_role_search,
                query_for_list=filter_roles_for_list,
                ordering=("name",),
            ),
            ModelConfig(
                model=DemoAuditORM,
                slug="audits",
                title="Журнал",
                list_display=("id", "role_id"),
                fields={
                    "role_id": AdminFieldConfig(
                        label="Роль",
                        relation_model=DemoRoleORM,
                        relation_label_field="name",
                    ),
                },
            ),
            ModelConfig(
                model=DemoPostORM,
                slug="posts",
                title="Посты",
                list_display=("id", "title"),
            ),
            ModelConfig(
                model=DemoCommentORM,
                slug="comments",
                title="Комментарии",
                list_display=("id", "body"),
            ),
            ModelConfig(
                model=DemoBadgeORM,
                slug="badges",
                title="Бейджи",
                list_display=("id", "title"),
            ),
            ModelConfig(
                model=DemoTaskORM,
                slug="tasks",
                title="Задачи",
                list_display=("id", "title"),
            ),
            ModelConfig(
                model=DemoOwnedProjectORM,
                slug="owned-projects",
                title="Projects",
                list_display=("id", "title"),
            ),
            ModelConfig(
                model=DemoOwnedBoardORM,
                slug="owned-boards",
                title="Boards",
                list_display=("id", "title"),
            ),
            ModelConfig(
                model=DemoTokenProfileORM,
                slug="token-profiles",
                title="Token Profiles",
                list_display=("tokenprofile_ptr_id", "title"),
            ),
            ModelConfig(
                model=DemoAliasedMessageORM,
                slug="aliased-messages",
                title="Aliased Messages",
                list_display=("id",),
            ),
            ModelConfig(
                model=DemoAliasedChatORM,
                slug="aliased-chats",
                title="Aliased Chats",
                list_display=("id", "last_branch_message_id"),
            ),
            ModelConfig(
                model=DemoSoftUserORM,
                slug="soft-users",
                title="Soft Users",
                list_display=("id", "username"),
            ),
            ModelConfig(
                model=DemoSoftNoteORM,
                slug="soft-notes",
                title="Soft Notes",
                list_display=("id", "body", "user_id"),
            ),
            ModelConfig(
                model=DemoHardParentORM,
                slug="hard-parents",
                title="Hard Parents",
                list_display=("id", "title"),
            ),
            ModelConfig(
                model=DemoHardChildORM,
                slug="hard-children",
                title="Hard Children",
                list_display=("id", "title", "parent_id"),
            ),
        ),
    )

    app = FastAPI()
    app.include_router(
        create_admin_router(
            AdminHTTPConfig(
                registry=config,
                get_db_session_dependency=get_session,
                get_current_user_dependency=lambda: SimpleNamespace(is_staff=is_staff),
                is_allowed=lambda user: bool(user.is_staff),
            ),
        ),
    )

    async with engine.begin() as connection:  # noqa
        await connection.run_sync(Base.metadata.create_all)

    return app, session_factory


@pytest.mark.asyncio
async def test_router_supports_virtual_fields_and_custom_setter() -> None:
    app, session_factory = await _build_app()
    transport = ASGITransport(app=app)

    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            create_response = await client.post(
                "/xladmin/models/users/items/",
                json={
                    "username": "tester",
                    "joined_on": "2026-03-26",
                    "new_password": "secret",
                },
            )

            assert create_response.status_code == 201
            body = create_response.json()
            assert body["item"]["new_password"] == ""

    async with session_factory() as session:
        item = (await session.execute(select(DemoUserORM))).scalar_one()
        assert item.password == "hashed::secret"


@pytest.mark.asyncio
async def test_router_uses_list_display_and_custom_search() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        session.add_all(
            [
                DemoUserORM(username="alpha", password="hashed::1"),
                DemoUserORM(username="beta", password="hashed::2"),
            ],
        )
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/", params={"q": "al"})

            assert response.status_code == 200
            payload = response.json()
            display_name_meta = next(
                field
                for field in payload["meta"]["fields"]
                if field["name"] == "display_name"
            )
            assert payload["meta"]["list_fields"] == ["id", "display_name", "joined_on", "is_active"]
            assert payload["meta"]["page_size"] == 120
            assert display_name_meta["width_px"] == 320
            assert payload["items"][0]["display_name"].endswith("alpha")
            assert payload["pagination"]["total"] == 1


@pytest.mark.asyncio
async def test_router_supports_custom_query_for_list() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        session.add_all(
            [
                DemoRoleORM(name="admin"),
                DemoRoleORM(name="hidden"),
            ],
        )
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/roles/items/")

            assert response.status_code == 200
            assert [item["name"] for item in response.json()["items"]] == ["admin"]


@pytest.mark.asyncio
async def test_router_returns_list_filters_in_model_meta() -> None:
    app, _session_factory = await _build_app()
    transport = ASGITransport(app=app)

    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/")

            assert response.status_code == 200
            assert response.json()["list_filters"] == [
                {
                    "slug": "status",
                    "label": "Status",
                    "group": "Flags",
                    "field_name": "is_active",
                    "input_kind": "boolean",
                    "placeholder": None,
                    "has_choices": True,
                    "options": [
                        {"value": "true", "label": "Да"},
                        {"value": "false", "label": "Нет"},
                    ],
                },
                {
                    "slug": "role_id",
                    "label": "Role",
                    "group": "Access",
                    "field_name": "roles",
                    "input_kind": "select",
                    "placeholder": None,
                    "has_choices": True,
                    "options": [],
                },
                {
                    "slug": "username_contains",
                    "label": "Username",
                    "group": None,
                    "field_name": "username",
                    "input_kind": "text",
                    "placeholder": None,
                    "has_choices": False,
                    "options": [],
                },
                {
                    "slug": "username_prefix",
                    "label": "Username Prefix",
                    "group": None,
                    "field_name": None,
                    "input_kind": "text",
                    "placeholder": None,
                    "has_choices": False,
                    "options": [],
                },
                {
                    "slug": "audience",
                    "label": "Audience",
                    "group": "Flags",
                    "field_name": None,
                    "input_kind": "select",
                    "placeholder": None,
                    "has_choices": True,
                    "options": [
                        {"value": "active-only", "label": "Active only"},
                        {"value": "inactive-only", "label": "Inactive only"},
                    ],
                },
            ]


@pytest.mark.asyncio
async def test_router_supports_boolean_list_filter() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        session.add_all(
            [
                DemoUserORM(username="alpha", password="hashed::1", is_active=True),
                DemoUserORM(username="beta", password="hashed::2", is_active=False),
            ],
        )
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/", params={"status": "false"})

            assert response.status_code == 200
            assert [item["display_name"] for item in response.json()["items"]] == ["Пользователь 2: beta"]


@pytest.mark.asyncio
async def test_router_returns_relation_filter_choices() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        session.add_all(
            [
                DemoRoleORM(name="admin"),
                DemoRoleORM(name="manager"),
            ],
        )
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/filters/role_id/choices/")

            assert response.status_code == 200
            assert response.json()["items"] == [
                {"id": 1, "label": "admin"},
                {"id": 2, "label": "manager"},
            ]


@pytest.mark.asyncio
async def test_router_supports_many_to_many_relation_list_filter() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        admin_role = DemoRoleORM(name="admin")
        manager_role = DemoRoleORM(name="manager")
        first_user = DemoUserORM(username="alpha", password="hashed::1", roles=[admin_role])
        second_user = DemoUserORM(username="beta", password="hashed::2", roles=[manager_role])
        session.add_all([first_user, second_user])
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/", params={"role_id": "1"})

            assert response.status_code == 200
            assert [item["display_name"] for item in response.json()["items"]] == ["Пользователь 1: alpha"]


@pytest.mark.asyncio
async def test_router_supports_text_list_filter() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        session.add_all(
            [
                DemoUserORM(username="alpha", password="hashed::1"),
                DemoUserORM(username="xlar", password="hashed::2"),
            ],
        )
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/", params={"username_contains": "xla"})

            assert response.status_code == 200
            assert [item["display_name"] for item in response.json()["items"]] == ["Пользователь 2: xlar"]


@pytest.mark.asyncio
async def test_router_supports_custom_list_filter_handler() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        session.add_all(
            [
                DemoUserORM(username="alpha", password="hashed::1"),
                DemoUserORM(username="xlar", password="hashed::2"),
            ],
        )
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/", params={"username_prefix": "xla"})

            assert response.status_code == 200
            assert [item["display_name"] for item in response.json()["items"]] == ["Пользователь 2: xlar"]


@pytest.mark.asyncio
async def test_router_supports_select_option_specific_filter_handler() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        session.add_all(
            [
                DemoUserORM(username="alpha", password="hashed::1", is_active=True),
                DemoUserORM(username="beta", password="hashed::2", is_active=False),
            ],
        )
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/", params={"audience": "inactive-only"})

            assert response.status_code == 200
            assert [item["display_name"] for item in response.json()["items"]] == ["Пользователь 2: beta"]


@pytest.mark.asyncio
async def test_router_supports_bulk_actions() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        session.add_all(
            [
                DemoUserORM(username="alpha", password="hashed::1", is_active=False),
                DemoUserORM(username="beta", password="hashed::2", is_active=False),
            ],
        )
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.post(
                "/xladmin/models/users/bulk-actions/activate/",
                json={"ids": [1, 2]},
            )

            assert response.status_code == 200
            assert response.json()["processed"] == 2
            assert response.json()["activated"] == 2

    async with session_factory() as session:
        items = list((await session.execute(select(DemoUserORM).order_by(DemoUserORM.id))).scalars())
        assert [item.is_active for item in items] == [True, True]


@pytest.mark.asyncio
async def test_router_reuses_related_model_search_for_choices() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        session.add_all(
            [
                DemoRoleORM(name="admin"),
                DemoRoleORM(name="moderator"),
                DemoRoleORM(name="user"),
            ],
        )
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get(
                "/xladmin/models/audits/fields/role_id/choices/",
                params={"q": "ad"},
            )

            assert response.status_code == 200
            assert response.json()["items"] == [{"id": 1, "label": "admin"}]


@pytest.mark.asyncio
async def test_router_supports_sorting_by_multiple_fields() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        session.add_all(
            [
                DemoUserORM(username="beta", password="hashed::1", joined_on=date(2026, 3, 27), is_active=False),
                DemoUserORM(username="alpha", password="hashed::2", joined_on=date(2026, 3, 25), is_active=True),
                DemoUserORM(username="gamma", password="hashed::3", joined_on=date(2026, 3, 26), is_active=False),
            ],
        )
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            by_display_response = await client.get(
                "/xladmin/models/users/items/",
                params={"sort": "display_name"},
            )
            by_multiple_fields_response = await client.get(
                "/xladmin/models/users/items/",
                params={"sort": "is_active,-id"},
            )
            by_date_response = await client.get(
                "/xladmin/models/users/items/",
                params={"sort": "joined_on"},
            )

            assert by_display_response.status_code == 200
            assert [item["id"] for item in by_display_response.json()["items"]] == [2, 1, 3]

            assert by_multiple_fields_response.status_code == 200
            assert [item["id"] for item in by_multiple_fields_response.json()["items"]] == [3, 1, 2]

            assert by_date_response.status_code == 200
            assert [item["joined_on"] for item in by_date_response.json()["items"]] == [
                "2026-03-25",
                "2026-03-26",
                "2026-03-27",
            ]


@pytest.mark.asyncio
async def test_router_supports_object_actions() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        session.add(DemoUserORM(username="alpha", password="hashed::1", is_active=True))
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            meta_response = await client.get("/xladmin/models/users/")
            action_response = await client.post("/xladmin/models/users/items/1/actions/deactivate/", json={})

            assert meta_response.status_code == 200
            assert meta_response.json()["object_actions"] == [
                {"slug": "deactivate", "label": "Деактивировать"},
            ]

            assert action_response.status_code == 200
            assert action_response.json()["result"]["deactivated"] == 1
            assert action_response.json()["item"]["is_active"] is False


@pytest.mark.asyncio
async def test_router_returns_english_locale_and_builtin_delete_label() -> None:
    app, _session_factory = await _build_app(locale="en")
    transport = ASGITransport(app=app)

    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/")

            assert response.status_code == 200
            payload = response.json()
            assert payload["locale"] == "en"
            assert payload["items"][0]["locale"] == "en"
            assert payload["items"][0]["bulk_actions"][0] == {"slug": "delete", "label": "Delete"}


@pytest.mark.asyncio
async def test_router_returns_locale_in_model_list_detail_and_item_meta() -> None:
    app, session_factory = await _build_app(locale="en")

    async with session_factory() as session:
        session.add(DemoUserORM(username="alpha", password="hashed::1"))
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            model_response = await client.get("/xladmin/models/users/")
            list_response = await client.get("/xladmin/models/users/items/")
            item_response = await client.get("/xladmin/models/users/items/1/")

            assert model_response.status_code == 200
            assert model_response.json()["locale"] == "en"

            assert list_response.status_code == 200
            assert list_response.json()["meta"]["locale"] == "en"

            assert item_response.status_code == 200
            assert item_response.json()["meta"]["locale"] == "en"


@pytest.mark.asyncio
async def test_router_builds_delete_preview_tree_for_object() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        user = DemoUserORM(username="alpha", password="hashed::1")
        post = DemoPostORM(title="Welcome", user=user)
        post.comments.append(DemoCommentORM(body="First"))
        session.add(user)
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/1/delete-preview/")

            assert response.status_code == 200
            payload = response.json()
            assert payload["can_delete"] is True
            assert payload["summary"] == {"roots": 1, "delete": 2, "protect": 0, "set_null": 0, "total": 3}
            assert payload["roots"][0]["model_slug"] == "users"
            assert payload["roots"][0]["children"][0]["model_slug"] == "posts"
            assert payload["roots"][0]["children"][0]["effect"] == "delete"
            assert payload["roots"][0]["children"][0]["children"][0]["model_slug"] == "comments"


@pytest.mark.asyncio
async def test_router_builds_delete_preview_tree_for_bulk_delete() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        first_user = DemoUserORM(username="alpha", password="hashed::1")
        first_post = DemoPostORM(title="Alpha post", user=first_user)
        first_post.comments.append(DemoCommentORM(body="Alpha comment"))

        second_user = DemoUserORM(username="beta", password="hashed::2")
        second_post = DemoPostORM(title="Beta post", user=second_user)
        second_post.comments.append(DemoCommentORM(body="Beta comment"))

        session.add_all([first_user, second_user])
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.post("/xladmin/models/users/bulk-delete-preview/", json={"ids": [1, 2]})

            assert response.status_code == 200
            payload = response.json()
            assert payload["can_delete"] is True
            assert payload["summary"] == {"roots": 2, "delete": 4, "protect": 0, "set_null": 0, "total": 6}
            assert [root["id"] for root in payload["roots"]] == [1, 2]


@pytest.mark.asyncio
async def test_router_returns_empty_bulk_delete_preview_for_empty_ids() -> None:
    app, _session_factory = await _build_app()
    transport = ASGITransport(app=app)

    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.post("/xladmin/models/users/bulk-delete-preview/", json={"ids": []})

            assert response.status_code == 200
            assert response.json() == {
                "can_delete": True,
                "summary": {"roots": 0, "delete": 0, "protect": 0, "set_null": 0, "total": 0},
                "roots": [],
            }


@pytest.mark.asyncio
async def test_router_builds_bulk_delete_preview_only_for_existing_ids() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        session.add(DemoUserORM(username="alpha", password="hashed::1"))
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.post("/xladmin/models/users/bulk-delete-preview/", json={"ids": [1, 999]})

            assert response.status_code == 200
            payload = response.json()
            assert payload["can_delete"] is True
            assert payload["summary"] == {"roots": 1, "delete": 0, "protect": 0, "set_null": 0, "total": 1}
            assert [root["id"] for root in payload["roots"]] == [1]


@pytest.mark.asyncio
async def test_router_marks_non_cascade_relations_as_protected_in_delete_preview() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        user = DemoUserORM(username="alpha", password="hashed::1")
        badge = DemoBadgeORM(title="VIP", user=user)
        session.add_all([user, badge])
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/1/delete-preview/")

            assert response.status_code == 200
            payload = response.json()
            assert payload["can_delete"] is False
            assert payload["summary"] == {"roots": 1, "delete": 0, "protect": 1, "set_null": 0, "total": 2}
            assert payload["roots"][0]["children"][0]["model_slug"] == "badges"
            assert payload["roots"][0]["children"][0]["effect"] == "protect"


@pytest.mark.asyncio
async def test_router_marks_required_non_cascade_relations_as_protected_in_delete_preview() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        user = DemoUserORM(username="alpha", password="hashed::1")
        user.tasks.append(DemoTaskORM(title="Follow-up"))
        session.add(user)
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/1/delete-preview/")

            assert response.status_code == 200
            payload = response.json()
            assert payload["can_delete"] is False
            assert payload["summary"] == {"roots": 1, "delete": 0, "protect": 1, "set_null": 0, "total": 2}
            assert payload["roots"][0]["children"][0]["model_slug"] == "tasks"
            assert payload["roots"][0]["children"][0]["effect"] == "protect"


@pytest.mark.asyncio
async def test_router_blocks_delete_for_required_non_cascade_relations() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        user = DemoUserORM(username="alpha", password="hashed::1")
        user.tasks.append(DemoTaskORM(title="Follow-up"))
        session.add(user)
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.delete("/xladmin/models/users/items/1/")

            assert response.status_code == 409

    async with session_factory() as session:
        assert len(list((await session.execute(select(DemoUserORM))).scalars())) == 1
        assert len(list((await session.execute(select(DemoTaskORM))).scalars())) == 1


@pytest.mark.asyncio
async def test_router_detects_reverse_foreign_keys_without_parent_relationships() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        user = DemoUserORM(username="owner", password="hashed::1")
        session.add(user)
        await session.flush()
        project = DemoOwnedProjectORM(title="Tracker Project", author_id=user.id)
        session.add(project)
        await session.flush()
        session.add(DemoOwnedBoardORM(title="Roadmap", project_id=project.id))
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/1/delete-preview/")

            assert response.status_code == 200
            payload = response.json()
            assert payload["can_delete"] is False
            assert payload["summary"] == {"roots": 1, "delete": 0, "protect": 2, "set_null": 0, "total": 3}
            project_node = payload["roots"][0]["children"][0]
            assert project_node["model_slug"] == "owned-projects"
            assert project_node["relation_name"] == "author_id"
            assert project_node["effect"] == "protect"
            assert project_node["children"][0]["model_slug"] == "owned-boards"
            assert project_node["children"][0]["relation_name"] == "project_id"
            assert project_node["children"][0]["effect"] == "protect"


@pytest.mark.asyncio
async def test_router_blocks_delete_for_reverse_foreign_keys_without_parent_relationships() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        user = DemoUserORM(username="owner", password="hashed::1")
        session.add(user)
        await session.flush()
        project = DemoOwnedProjectORM(title="Tracker Project", author_id=user.id)
        session.add(project)
        await session.flush()
        session.add(DemoOwnedBoardORM(title="Roadmap", project_id=project.id))
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.delete("/xladmin/models/users/items/1/")

            assert response.status_code == 409

    async with session_factory() as session:
        assert len(list((await session.execute(select(DemoUserORM))).scalars())) == 1
        assert len(list((await session.execute(select(DemoOwnedProjectORM))).scalars())) == 1
        assert len(list((await session.execute(select(DemoOwnedBoardORM))).scalars())) == 1


@pytest.mark.asyncio
async def test_router_handles_related_models_with_non_default_primary_key() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        user = DemoUserORM(username="owner", password="hashed::1")
        session.add(user)
        await session.flush()
        session.add(
            DemoTokenProfileORM(
                tokenprofile_ptr_id="profile-1",
                title="Profile",
                user_id=user.id,
            ),
        )
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/1/delete-preview/")

            assert response.status_code == 200
            payload = response.json()
            assert payload["can_delete"] is False
            related_slugs = [child["model_slug"] for child in payload["roots"][0]["children"]]
            assert "token-profiles" in related_slugs


@pytest.mark.asyncio
async def test_router_handles_related_models_with_aliased_column_name() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        message = DemoAliasedMessageORM(id="message-1")
        session.add(message)
        await session.flush()
        session.add(DemoAliasedChatORM(id="chat-1", last_branch_message_id=message.id))
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/aliased-messages/items/message-1/delete-preview/")

            assert response.status_code == 200
            payload = response.json()
            assert payload["can_delete"] is False
            assert payload["summary"] == {"roots": 1, "delete": 0, "protect": 1, "set_null": 0, "total": 2}
            assert payload["roots"][0]["children"][0]["model_slug"] == "aliased-chats"


@pytest.mark.asyncio
async def test_router_marks_ondelete_set_null_relations_in_delete_preview() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        user = DemoSoftUserORM(username="alpha")
        session.add(user)
        await session.flush()
        session.add(DemoSoftNoteORM(body="note", user_id=user.id))
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/soft-users/items/1/delete-preview/")

            assert response.status_code == 200
            payload = response.json()
            assert payload["can_delete"] is True
            assert payload["summary"] == {"roots": 1, "delete": 0, "protect": 0, "set_null": 1, "total": 2}
            assert payload["roots"][0]["children"][0]["model_slug"] == "soft-notes"
            assert payload["roots"][0]["children"][0]["effect"] == "set-null"


@pytest.mark.asyncio
async def test_router_applies_set_null_before_delete() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        user = DemoSoftUserORM(username="alpha")
        session.add(user)
        await session.flush()
        session.add(DemoSoftNoteORM(body="note", user_id=user.id))
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.delete("/xladmin/models/soft-users/items/1/")

            assert response.status_code == 204

    async with session_factory() as session:
        assert len(list((await session.execute(select(DemoSoftUserORM))).scalars())) == 0
        note = (await session.execute(select(DemoSoftNoteORM))).scalar_one()
        assert note.user_id is None


@pytest.mark.asyncio
async def test_router_marks_ondelete_cascade_relations_as_delete() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        parent = DemoHardParentORM(title="parent")
        session.add(parent)
        await session.flush()
        session.add(DemoHardChildORM(title="child", parent_id=parent.id))
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/hard-parents/items/1/delete-preview/")

            assert response.status_code == 200
            payload = response.json()
            assert payload["can_delete"] is True
            assert payload["summary"] == {"roots": 1, "delete": 1, "protect": 0, "set_null": 0, "total": 2}
            assert payload["roots"][0]["children"][0]["model_slug"] == "hard-children"
            assert payload["roots"][0]["children"][0]["effect"] == "delete"


@pytest.mark.asyncio
async def test_router_deletes_ondelete_cascade_relations() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        parent = DemoHardParentORM(title="parent")
        session.add(parent)
        await session.flush()
        session.add(DemoHardChildORM(title="child", parent_id=parent.id))
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.delete("/xladmin/models/hard-parents/items/1/")

            assert response.status_code == 204

    async with session_factory() as session:
        assert len(list((await session.execute(select(DemoHardParentORM))).scalars())) == 0
        assert len(list((await session.execute(select(DemoHardChildORM))).scalars())) == 0


@pytest.mark.asyncio
async def test_router_builds_delete_preview_with_multiple_cascade_branches() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        user = DemoUserORM(username="alpha", password="hashed::1")
        post = DemoPostORM(title="Welcome", user=user)
        post.comments.append(DemoCommentORM(body="First"))
        user.secrets.append(DemoSecretORM(code_name="vault"))
        session.add(user)
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/1/delete-preview/")

            assert response.status_code == 200
            payload = response.json()
            assert payload["can_delete"] is True
            assert payload["summary"] == {"roots": 1, "delete": 3, "protect": 0, "set_null": 0, "total": 4}
            child_model_slugs = [child["model_slug"] for child in payload["roots"][0]["children"]]
            assert "posts" in child_model_slugs
            assert None in child_model_slugs


@pytest.mark.asyncio
async def test_router_uses_fallback_label_for_unregistered_related_model_in_delete_preview() -> None:
    app, session_factory = await _build_app()

    async with session_factory() as session:
        user = DemoUserORM(username="alpha", password="hashed::1")
        user.secrets.append(DemoSecretORM(code_name="vault"))
        session.add(user)
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/1/delete-preview/")

            assert response.status_code == 200
            payload = response.json()
            secret_node = payload["roots"][0]["children"][0]
            assert secret_node["model_slug"] is None
            assert secret_node["model_title"] == "DemoSecretORM"
            assert secret_node["label"] == "vault"


@pytest.mark.asyncio
async def test_router_returns_localized_forbidden_error() -> None:
    app, _session_factory = await _build_app(locale="en", is_staff=False)
    transport = ASGITransport(app=app)

    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/")

            assert response.status_code == 403
            assert response.json()["detail"] == "Access denied."


@pytest.mark.asyncio
async def test_router_returns_localized_object_not_found_error() -> None:
    app, _session_factory = await _build_app(locale="en")
    transport = ASGITransport(app=app)

    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/users/items/999/delete-preview/")

            assert response.status_code == 404
            assert response.json()["detail"] == "Object was not found."


@pytest.mark.asyncio
async def test_router_returns_localized_admin_model_not_found_error() -> None:
    app, _session_factory = await _build_app(locale="en")
    transport = ASGITransport(app=app)

    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.get("/xladmin/models/missing-model/")

            assert response.status_code == 404
            assert response.json()["detail"] == "Admin model was not found."


@pytest.mark.asyncio
async def test_router_returns_localized_object_action_not_found_error() -> None:
    app, session_factory = await _build_app(locale="en")

    async with session_factory() as session:
        session.add(DemoUserORM(username="alpha", password="hashed::1"))
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.post("/xladmin/models/users/items/1/actions/missing-action/", json={})

            assert response.status_code == 404
            assert response.json()["detail"] == "Object action was not found."


@pytest.mark.asyncio
async def test_router_returns_localized_bulk_action_not_found_error() -> None:
    app, session_factory = await _build_app(locale="en")

    async with session_factory() as session:
        session.add(DemoUserORM(username="alpha", password="hashed::1"))
        await session.commit()

    transport = ASGITransport(app=app)
    async with transport:
        async with AsyncClient(transport=transport, base_url="http://testserver") as client:
            response = await client.post("/xladmin/models/users/bulk-actions/missing-action/", json={"ids": [1]})

            assert response.status_code == 404
            assert response.json()["detail"] == "Bulk action was not found."
